# the following method contains portions of the software mdtraj which
# is distributed under the following license
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Peter Eastman, Robert McGibbon
# Contributors: Kyle A. Beauchamp, Matthew Harrigan, Carlos Xavier Hernandez
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit, copyright (c) 2012 Stanford University and Peter Eastman. Those
# portions are distributed under the following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################


# These van der waals radii are taken from
# https://en.wikipedia.org/wiki/Atomic_radii_of_the_elements_(data_page)
# which references
# A. Bondi (1964). "van der Waals Volumes and Radii". J. Phys. Chem. 68: 441. doi:10.1021/j100785a001 and doi:10.1021/jp8111556.
# M. Mantina et al. (2009). "Consistent van der Waals Radii for the Whole Main Group". J. Phys. Chem. A. 113 (19): 5806--12. doi:10.1021/jp8111556
# Where no van der Waals value is known, a default of 2 angstroms is used.
# However, because certain atoms in biophysical simulations have a high
# chance of being completely ionized, we have decided to give the
# following atoms their ionic radii their ionic radii:
# +2: Be, Mg, Ca, Ba
# +1: Li, Na, K, Cs
# -1: Cl
# These ionic radii are were taken from:
# Shannon, R. D. Revised effective ionic radii and systematic studies of interatomic distances in halides and chalcogenides. Acta Crystallographica Section A 32, 751--767 (1976). doi:10.1107/S0567739476001551
# For most atoms, adding electrons usually doesn't change the radius much
# (<10%), while removing them changes it substantially (>50%). Further,
# when atoms like N, S, and P, are positive, they are bound to atoms in such
# a way that would "hide" their radii anyway. We have therefore chosen to just
# use their vdW radii.

ATOMIC_RADII = (('H', 0.12),
                ('He', 0.14),
                ('Li', 0.076),
                ('Be', 0.059),
                ('B', 0.192),
                ('C', 0.17),
                ('N', 0.155),
                ('O', 0.152),
                ('F', 0.147),
                ('Ne', 0.154),
                ('Na', 0.102),
                ('Mg', 0.086),
                ('Al', 0.184),
                ('Si', 0.21),
                ('P', 0.18),
                ('S', 0.18),
                ('Cl', 0.181),
                ('Ar', 0.188),
                ('K', 0.138),
                ('Ca', 0.114),
                ('Sc', 0.211),
                ('Ti', 0.2),
                ('V', 0.2),
                ('Cr', 0.2),
                ('Mn', 0.2),
                ('Fe', 0.2),
                ('Co', 0.2),
                ('Ni', 0.163),
                ('Cu', 0.14),
                ('Zn', 0.139),
                ('Ga', 0.187),
                ('Ge', 0.211),
                ('As', 0.185),
                ('Se', 0.19),
                ('Br', 0.185),
                ('Kr', 0.202),
                ('Rb', 0.303),
                ('Sr', 0.249),
                ('Y', 0.2),
                ('Zr', 0.2),
                ('Nb', 0.2),
                ('Mo', 0.2),
                ('Tc', 0.2),
                ('Ru', 0.2),
                ('Rh', 0.2),
                ('Pd', 0.163),
                ('Ag', 0.172),
                ('Cd', 0.158),
                ('In', 0.193),
                ('Sn', 0.217),
                ('Sb', 0.206),
                ('Te', 0.206),
                ('I', 0.198),
                ('Xe', 0.216),
                ('Cs', 0.167),
                ('Ba', 0.149),
                ('La', 0.2),
                ('Ce', 0.2),
                ('Pr', 0.2),
                ('Nd', 0.2),
                ('Pm', 0.2),
                ('Sm', 0.2),
                ('Eu', 0.2),
                ('Gd', 0.2),
                ('Tb', 0.2),
                ('Dy', 0.2),
                ('Ho', 0.2),
                ('Er', 0.2),
                ('Tm', 0.2),
                ('Yb', 0.2),
                ('Lu', 0.2),
                ('Hf', 0.2),
                ('Ta', 0.2),
                ('W', 0.2),
                ('Re', 0.2),
                ('Os', 0.2),
                ('Ir', 0.2),
                ('Pt', 0.175),
                ('Au', 0.166),
                ('Hg', 0.155),
                ('Tl', 0.196),
                ('Pb', 0.202),
                ('Bi', 0.207),
                ('Po', 0.197),
                ('At', 0.202),
                ('Rn', 0.22),
                ('Fr', 0.348),
                ('Ra', 0.283),
                ('Ac', 0.2),
                ('Th', 0.2),
                ('Pa', 0.2),
                ('U', 0.186),
                ('Np', 0.2),
                ('Pu', 0.2),
                ('Am', 0.2),
                ('Cm', 0.2),
                ('Bk', 0.2),
                ('Cf', 0.2),
                ('Es', 0.2),
                ('Fm', 0.2),
                ('Md', 0.2),
                ('No', 0.2),
                ('Lr', 0.2),
                ('Rf', 0.2),
                ('Db', 0.2),
                ('Sg', 0.2),
                ('Bh', 0.2),
                ('Hs', 0.2),
                ('Mt', 0.2),
                ('Ds', 0.2),
                ('Rg', 0.2),
                ('Cn', 0.2),
                ('Uut', 0.2),
                ('Fl', 0.2),
                ('Uup', 0.2),
                ('Lv', 0.2),
                ('Uus', 0.2),
                ('Uuo', 0.2))





import numpy as np
import pytest
from geomm.grouping import group_pair

def test_group_pair_no_shift():
    # member_b centroid is close to member_a, no shift needed
    coords = np.array([
        [1.0, 1.0, 1.0],  # member_a
        [2.0, 2.0, 2.0],  # member_b
    ])
    unitcell = np.array([10.0, 10.0, 10.0])
    member_a_idxs = [0]
    member_b_idxs = [1]
    result = group_pair(coords, unitcell, member_a_idxs, member_b_idxs)
    np.testing.assert_allclose(result, coords)

def test_group_pair_positive_shift():
    # member_b centroid is too low in x, should be shifted by +unitcell[0]
    coords = np.array([
        [1.0, 1.0, 1.0],   # member_a
        [-8.0, 1.0, 1.0],   # member_b
    ])
    unitcell = np.array([10.0, 10.0, 10.0])
    member_a_idxs = [0]
    member_b_idxs = [1]
    expected = np.array([
        [1.0, 1.0, 1.0],
        [2.0, 1.0, 1.0],  # shifted by +10 in x
    ])
    result = group_pair(coords, unitcell, member_a_idxs, member_b_idxs)
    np.testing.assert_allclose(result, expected)

def test_group_pair_negative_shift():
    # member_b centroid is to high in -y, should be shifted by -unitcell[1]
    coords = np.array([
        [1.0, 0.0, 1.0],   # member_a
        [1.0, 8.0, 1.0],   # member_b
    ])
    unitcell = np.array([10.0, 10.0, 10.0])
    member_a_idxs = [0]
    member_b_idxs = [1]
    expected = np.array([
        [1.0, 0.0, 1.0],
        [1.0, -2.0, 1.0],  # shifted by -10 in y
    ])
    result = group_pair(coords, unitcell, member_a_idxs, member_b_idxs)
    np.testing.assert_allclose(result, expected)

def test_group_pair_shift_multiple_dims():
    # member_b centroid is far in +x and -z, should be shifted in both
    coords = np.array([
        [1.0, 1.0, 8.0],   # member_a
        [8.0, 1.0, 1.0],   # member_b
    ])
    unitcell = np.array([10.0, 10.0, 10.0])
    member_a_idxs = [0]
    member_b_idxs = [1]
    expected = np.array([
        [1.0, 1.0, 8.0],
        [-2.0, 1.0, 11.0],  # shifted by -10 in x, +10 in z
    ])
    result = group_pair(coords, unitcell, member_a_idxs, member_b_idxs)
    np.testing.assert_allclose(result, expected)

def test_group_pair_multiple_atoms_per_group():
    # Test with multiple atoms per group
    coords = np.array([
        [1.0, 1.0, 1.0],   # member_a
        [1.0, 2.0, 1.0],   # member_a
        [8.0, 1.0, 1.0],   # member_b
        [8.0, 2.0, 1.0],   # member_b
    ])
    unitcell = np.array([10.0, 10.0, 10.0])
    member_a_idxs = [0, 1]
    member_b_idxs = [2, 3]
    expected = np.array([
        [1.0, 1.0, 1.0],
        [1.0, 2.0, 1.0],
        [-2.0, 1.0, 1.0],  # shifted by -10 in x
        [-2.0, 2.0, 1.0],  # shifted by -10 in x
    ])
    result = group_pair(coords, unitcell, member_a_idxs, member_b_idxs)
    np.testing.assert_allclose(result, expected)


# the following method contains portions of the software mdtraj which
# is distributed under the following license
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Peter Eastman, Robert McGibbon
# Contributors: Kyle A. Beauchamp, Matthew Harrigan, Carlos Xavier Hernandez
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit, copyright (c) 2012 Stanford University and Peter Eastman. Those
# portions are distributed under the following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################

def shrake_rupley(coords, radii, probe_radius=0.14, n_sphere_points=960):
    """


    Notes
    -----
    This code implements the Shrake and Rupley algorithm, with the Golden
    Section Spiral algorithm to generate the sphere points. The basic idea
    is to great a mesh of points representing the surface of each atom
    (at a distance of the van der waals radius plus the probe
    radius from the nuclei), and then count the number of such mesh points
    that are on the molecular surface -- i.e. not within the radius of another
    atom. Assuming that the points are evenly distributed, the number of points
    is directly proportional to the accessible surface area (its just 4*pi*r^2
    time the fraction of the points that are accessible).
    There are a number of different ways to generate the points on the sphere --
    possibly the best way would be to do a little "molecular dyanmics" : put the
    points on the sphere, and then run MD where all the points repel one another
    and wait for them to get to an energy minimum. But that sounds expensive.
    This code uses the golden section spiral algorithm
    (picture at http://xsisupport.com/2012/02/25/evenly-distributing-points-on-a-sphere-with-the-golden-sectionspiral/)
    where you make this spiral that traces out the unit sphere and then put points
    down equidistant along the spiral. It's cheap, but not perfect.
    The gromacs utility g_sas uses a slightly different algorithm for generating
    points on the sphere, which is based on an icosahedral tesselation.
    roughly, the icosahedral tesselation works something like this
    http://www.ziyan.info/2008/11/sphere-tessellation-using-icosahedron.html
    References
    ----------
    .. [1] Shrake, A; Rupley, JA. (1973) J Mol Biol 79 (2): 351--71.

    """

    # the code needs to be moved over from mdtraj in a distilled form

    # currently it is implemented in C++ and the build process for
    # that needs to be figured out
    raise NotImplementedError

    pass



import numpy as np

def free_energy(weights, max_energy=None, zero_point_energy=1.0e-12,
                supermax_value=np.nan):
    """Perform a transformation of probability weights to free energies.

    The free energy of probabilities is a transformation of -ln(p)
    where p is the properly normalized probability.

    In addition to this simple transformation we set a maximum free
    energy (max_energy) and a mask value (supermax_value) for values
    greater than the max energy. This is to account for missing
    observations or extreme outliers. This can be disabled by setting
    max_energy to None.

    Free energy values are relative to a zero-point energy, which in
    the absence of a theoretically determined value can be simply
    determined by the minimum value in the dataset. If the
    zero_point_energy value is given this simply fixes this minimum
    free energy to the zero-point energy so that the rest of the
    values have some meaningful value. Typically, we set this to
    something to close to but greater than 0. Because a 0 free energy
    is equivalent to a probability of 1.0, which would make the
    dataset non-normalized.

    Parameters
    ----------

    weights : arraylike of float
        The normalized weights to transform

    max_energy : float or None
        Sets a maximum value for free energies. Free energies larger
        than this will be replaced by the 'supermax_value'. If None
        the masking is not performed.
       (Default = None)

    supermax_value : numpy scalar
        The value that replaces free energies that are larger than the
        max energy. Typically is a numpy NaN so as to omit the data.
       (Default = np.nan)

    zero_point_energy : float or None
        Free energies are relative and the minimum value will be set
        to this zero point energy. This is to avoid having 0 free
        energy values. If None this is not done.
       (Default = 1.0e-12)

    Returns
    -------

    free_energies : arraylike
       The free energy for each weight.

    """

    # assume the weights are normalized
    # -log(weights)
    free_energies = np.negative(np.log(weights))

    if zero_point_energy is not None:
        # set the min as the 0 energy
        free_energies = free_energies - free_energies.min()

        # then increase everything by the zero-point energy so there are
        # no zero energy states
        free_energies += zero_point_energy

    # energies greater than the max are set to the supermax_value
    if max_energy is not None:
        free_energies[np.where(free_energies > max_energy)] = supermax_value

    return free_energies



import warnings

import numpy as np

def box_vectors_to_lengths_angles(box_vectors):
    """Convert box vectors for a single 'frame' to lengths and angles."""

    # calculate the lengths of the vectors through taking the norm of
    # them
    unitcell_lengths = []
    for basis in box_vectors:
        unitcell_lengths.append(np.linalg.norm(basis))
    unitcell_lengths = np.array(unitcell_lengths)

    # calculate the angles for the vectors
    unitcell_angles = np.array([np.degrees(
                        np.arccos(np.dot(box_vectors[i], box_vectors[j])/
                                  (np.linalg.norm(box_vectors[i]) * np.linalg.norm(box_vectors[j]))))
                       for i, j in [(0,1), (1,2), (2,0)]])

    return unitcell_lengths, unitcell_angles


# License applicable to the function 'lengths_and_angles_to_box_vectors'
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

def lengths_and_angles_to_box_vectors(lengths, angles):
    """Convert from the lengths/angles of the unit cell to the box
    vectors (Bravais vectors). The angles should be in degrees.

    Parameters
    ----------
    lengths : arraylike of dim 3
        The lengths of each box vector

    angles : arraylike of dim 3
        The angles between each box vector, in degrees.

    Returns
    -------
    box_vectors : arraylike of shape (3, 3)
        If the inputs are scalar, the vectors will one dimensional (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)

    Examples
    --------

    >>> import numpy as np
    >>> result = lengths_and_angles_to_box_vectors(1, 1, 1, 90.0, 90.0, 90.0)

    Notes
    -----
    This code is adapted from gyroid, which is licensed under the BSD
    http://pythonhosted.org/gyroid/_modules/gyroid/unitcell.html

    """

    a_length, b_length, c_length = lengths
    alpha, beta, gamma = angles

    if np.all(alpha < 2*np.pi) and np.all(beta < 2*np.pi) and np.all(gamma < 2*np.pi):
        warnings.warn('All your angles were less than 2*pi. Did you accidentally give me radians?')

    alpha = alpha * np.pi / 180
    beta = beta * np.pi / 180
    gamma = gamma * np.pi / 180

    a = np.array([a_length, np.zeros_like(a_length), np.zeros_like(a_length)])
    b = np.array([b_length*np.cos(gamma), b_length*np.sin(gamma), np.zeros_like(b_length)])
    cx = c_length*np.cos(beta)
    cy = c_length*(np.cos(alpha) - np.cos(beta)*np.cos(gamma)) / np.sin(gamma)
    cz = np.sqrt(c_length*c_length - cx*cx - cy*cy)
    c = np.array([cx,cy,cz])

    if not a.shape == b.shape == c.shape:
        raise TypeError('Shape is messed up.')

    # Make sure that all vector components that are _almost_ 0 are set exactly
    # to 0
    tol = 1e-6
    a[np.logical_and(a>-tol, a<tol)] = 0.0
    b[np.logical_and(b>-tol, b<tol)] = 0.0
    c[np.logical_and(c>-tol, c<tol)] = 0.0

    return a.T, b.T, c.T



import numpy as np

import pint

unit = pint.UnitRegistry()

BOLTZMANN_CONSTANT = 1.3806504e-23 * (unit.joule / unit.kelvin)
PLANCK_CONSTANT = 6.626070e-34 * (unit.joule * unit.second)
UNIVERSAL_GAS_CONSTANT = 8.314 * (unit.joule / (unit.kelvin * unit.mole))
TRANSMISSION_COEFFICIENT = 1.0

def arrhenius_preexponent(temperature):

    preexponent = (TRANSMISSION_COEFFICIENT * BOLTZMANN_CONSTANT * temperature) /\
                  PLANCK_CONSTANT

    return preexponent

def arrhenius_rate(activation_free_energy, temperature):

    preexponent = arrhenius_preexponent(temperature)

    exponent = -1 * (activation_free_energy / (UNIVERSAL_GAS_CONSTANT * temperature))

    rate = preexponent * np.exp(exponent)

    return rate

def arrhenius_activation_free_energy(rate, temperature):

    preexponent = arrhenius_preexponent(temperature)

    free_energy = -(UNIVERSAL_GAS_CONSTANT * temperature) * np.log(rate/preexponent)

    return free_energy



import numpy as np

from geomm.centering import center

def group_complex(coords, complexes_idxs):
    """TODO
    """

    raise NotImplementedError

    grouped_coords = np.copy(coords)

    # get the coordinates of each complex
    complex_coords = []
    for complex_idxs in complexes_idxs:
        complex_coords.append(coords[complex_idxs, :])

    return grouped_coords

def group_pair(coords, unitcell_side_lengths, member_a_idxs, member_b_idxs):
    """For a pair of group of coordinates (e.g. atoms) this moves member_b
    coordinates to the image of the periodic unitcell that minimizes
    the difference between the centers of geometry between the two
    members (e.g. a protein and ligand).

    Parameters
    ----------

    coords : arraylike
        The coordinate array of the particles you will be
        transforming.

    unitcell_side_lengths : arraylike of shape (3)
        The lengths of the sides of a rectangular unitcell.

    member_a_idxs : arraylike of int of rank 1
        Collection of the indices that define that member of the pair.

    member_b_idxs : arraylike of int of rank 1
        Collection of the indices that define that member of the pair.

    Returns
    -------

    grouped_coords : arraylike
        Transformed coordinates.


    """

    # take the difference between the average coords of each
    # molecule.
    unitcell_half_lengths = unitcell_side_lengths * 0.5

    # initialize a new array for the coordinates
    grouped_coords = np.copy(coords)

    # get the coordinates for the ligand and member_b separately
    member_a_coords = coords[member_a_idxs, :]
    member_b_coords = coords[member_b_idxs, :]

    # calculate the centroids for each member
    member_a_centroid = member_a_coords.mean(axis=0)
    member_b_centroid = member_b_coords.mean(axis=0)

    # calculate the difference between them from the given coordinates
    centroid_dist = member_a_centroid - member_b_centroid

    # now we need to move one relative to the other by unitcell
    # lengths in order to find one that makes the difference between
    # the centroids smallest

    # When the centroid_dist are larger than the unitcell half length in
    # both the positive and negative direction

    # The positive direction

    # this gives us a tuple of two arrays. THe first array is the
    # indices of the frames that satisfied the boolean expression. The
    # second array is the index of the dimension that satisfied the
    # boolean expression, i.e. 0,1,2 for x,y,z. For example if we have
    # a frame where member_b is outside the box on two dimensions we
    # could have (array([0,0]), array([0,2])) for frame 1 on both x
    # and z dimensions. and only (array([0]), array([0])) for outside
    # on the first frame x dimension.
    pos_idxs = np.where(centroid_dist > unitcell_half_lengths)[0]

    # we simply pair elements from each of those arrays (pairwise) to
    # iterate over them. Essentially reshaping the two arrays
    for dim_idx in pos_idxs:
        # change the ligand coordinates to center them by adding the
        # cube side length in that direction
        grouped_coords[member_b_idxs, dim_idx] = (member_b_coords[:, dim_idx] +
                                                unitcell_side_lengths[dim_idx])

    # the negative direction is the same as the positive direction
    # except for the boolean expression in the where command and we
    # subtract the cube side length instead of adding it
    neg_idxs = np.where(centroid_dist < -unitcell_half_lengths)[0]
    for dim_idx in neg_idxs:
        grouped_coords[member_b_idxs, dim_idx] = (member_b_coords[:, dim_idx] -
                                                  unitcell_side_lengths[dim_idx])


    return grouped_coords

def shorten_vec(x, unitcell_side_lengths):
    """
    For a given vector x between two points in a periodic box, return the
    shortest version of that vector.
    """
    pos_idxs = np.where(x > 0.5*unitcell_side_lengths)[0]

    for dim_idx in pos_idxs:
        x[dim_idx] += unitcell_side_lengths[dim_idx]

    neg_idxs = np.where(x < -0.5*unitcell_side_lengths)[0]

    for dim_idx in neg_idxs:
        x[dim_idx] -= unitcell_side_lengths[dim_idx]

    return x


import numpy as np
import scipy.spatial.distance as dist
from scipy.spatial import KDTree

def distance():
    return None

def minimum_distance(coordsA, coordsB):
    """Calculate the minimum distance between members of coordsA and coordsB.
    Uses a fast binary search algorithm of order N*log(N).

    Parameters
    ----------

    coordsA : arraylike, shape (Natoms_A, 3)
        First set of coordinates.

    coordsB : arraylike, shape (Natoms_B, 3)
        Second set of coordinates.

    """

    # make sure the number of dimensions is 3
    assert (coordsA.shape[1] == 3) and (coordsB.shape[1] == 3), \
        "Minimum distance expecting arrays of shape (N, 3)"

    tree = KDTree(coordsA)
    return(tree.query(coordsB)[0].min())



"""Methods for calculating temperatures from the velocities of atoms."""

import numpy as np

BOLTZMANN_CONSTANT = 1.3806504e-23

def kinetic_energy(velocities, masses):

    return 0.5 * sum(masses * np.array([np.dot(vel, vel) for vel in velocities]))

def temperature(velocities, masses):

    return (2. / (3. * BOLTZMANN_CONSTANT)) * kinetic_energy(velocities, masses)



import numpy as np

from geomm.centroid import centroid

def center(coords, center_point):
    """Center coordinates at the origin based on a center point.

    If idxs are given the center of mass is computed only from those
    coordinates and if weights are given a weighted center of mass is
    computed.

    Parameters
    ----------

    coords : arraylike
       The coordinates you wish to center.

    center_point : arraylike of float of 1 dimension
       The point to center all other coordinates around. Should have
       one element for each dimension of coords

    Returns
    -------

    centered_coords : arraylike
        Transformed coordinates.

    """

    assert len(coords.shape) == 2, \
        "coordinates should be rank 2 array, "\
        "this function operates on individual frames not trajectories."
    assert coords.shape[1] == 3, "coordinates are not of 3 dimensions"
    assert len(center_point) == 3, "center point is not of 3 dimensions"

    return coords - center_point

def center_around(coords, idxs, weights=None):
    """Center coordinates at the origin based on a center point.

    If idxs are given the center of mass is computed only from those
    coordinates and if weights are given a weighted center of mass is
    computed.

    Parameters
    ----------

    coords : arraylike
       The coordinates you wish to center.

    idxs : arraylike of int
        The idxs of the coordinates to actually compute the centroid
        on, although the translation will act on all the coordinates.

    weights : arraylike of float, optional
        Give weights to the coordinates for a weighted centroid
        ('center of mass').

    Returns
    -------

    centered_coords : arraylike
        Transformed coordinates.

    """

    assert len(coords.shape) == 2, \
        "coordinates should be rank 2 array, "\
        "this function operates on individual frames not trajectories."
    assert coords.shape[1] == 3, "coordinates are not of 3 dimensions"
    assert len(idxs) > 0, "Must provide some idxs to compute a center of."

    return center(coords, centroid(coords[idxs], weights=weights))


def apply_rectangular_pbcs(coords, unitcell_side_lengths, center_point=(0., 0., 0.,)):
    """Apply rectangular Periodic Boundary Conditions (PBCs) given the
    lengths of the unitcell and a center point positions of the box in
    the coordinate space. The default for the center point is (0,0,0)
    which is the case for OpenMM MD frames but not other MD systems.

    Parameters
    ----------

    coords : arraylike
        The coordinate array of the particles you will be
        transforming.

    unitcell_side_lengths : arraylike of shape (3)
        The lengths of the sides of a rectangular unitcell.

    Returns
    -------

    wrapped_coords : arraylike
        Transformed coordinates. All fit within the box.

    Warning
    -------

    This method does not understand molecular topologies and will
    "break" bonds when moving molecules through boundaries.

    """

    # check to make sure everything looks okay
    assert len(coords.shape) == 2, \
        "coordinates should be rank 2 array, "\
        "this function operates on individual frames not trajectories."
    assert coords.shape[1] == 3, "coordinates are not of 3 dimensions"
    assert len(center_point) == 3, "center point is not of 3 dimensions"
    assert len(unitcell_side_lengths) == 3, "Unitcell side lengths are not of dimension 3"

    # cast the center point to an array
    center_point = np.array(center_point)

    # Calculate half box sizes
    unitcell_half_lengths = unitcell_side_lengths * 0.5

    # initialize the coordinates to be wrapped
    wrapped_coords = np.copy(coords)

    # find coords which are outside the box in the positive direction
    pos_idxs = np.where(coords > center_point + unitcell_half_lengths)

    # Groups the frame_idx, atom_idx, and dim_idx
    pos_idxs = list(zip(pos_idxs[0], pos_idxs[1]))

    # Restrict particle coordinates to the simulation box
    for atom_idx, dim_idx in pos_idxs:
        wrapped_coords[atom_idx, dim_idx] = (coords[atom_idx, dim_idx] -
                                                 unitcell_side_lengths[dim_idx])

    # Find where coords are less than  the negative half box sizes
    neg_idxs = np.where(coords < center_point - unitcell_half_lengths)

    # Groups the fram_idx, atom_idx and dim_idx where they are greater
    # than half box sizes
    neg_idxs = list(zip(neg_idxs[0], neg_idxs[1]))

    # Restrict particle coordinates to the simulation box
    for atom_idx, dim_idx in neg_idxs:
        wrapped_coords[atom_idx, dim_idx] = (coords[atom_idx, dim_idx] +
                                                 unitcell_side_lengths[dim_idx])

    return wrapped_coords


def center_complex(coords, complex_idxs):
    """For a system with periodic boundary conditions move all members of
    a complex to the same image of the unitcell.

    Parameters
    ----------

    coords : arraylike
        The coordinate array of the particles you will be
        transforming.

    complex_idxs : list of arraylikes of int of rank 1
        A list where each member represents a member of the complex
        and is a collection of the indices that define that member.

    Returns
    -------

    centered_coords : arraylike
        Transformed coordinates.

    """

    # compute the centroids of each member in the complex
    member_centroids = []
    for member_idxs in complex_idxs:
        centroid = coords[member_idxs].mean(axis=0)
        member_centroids.append(centroid)
    member_centroids = np.array(member_centroids)

    # compute the centroid of the centroids
    complex_centroid = member_centroids.mean(axis=0)

    # center the complex
    centered_coords = center(coords, complex_centroid)

    return centered_coords






# make sure to install dependencies in 'conversion.requirements.txt'
import os
import os.path as osp
import shutil

import sqlite3

import numpy as np
import pandas as pd
import tables as pytables
import mdtraj as mdj
from wepy.util.mdtraj import mdtraj_to_json_topology
from wepy.util.json_top import json_top_chain_df, json_top_residue_df, json_top_atom_df

traj = mdj.load('../lysozyme_pxylene.pdb')

if osp.exists("outputs"):
    shutil.rmtree("outputs")
os.makedirs("outputs")


# the JSON format
json_top = mdtraj_to_json_topology(traj.top)

with open('outputs/lysozyme_pxylene.top.json', 'w') as wf:
    wf.write(json_top)

# FASTA residue sequence
fasta_str = traj.top.to_fasta(chain=0)
with open('outputs/lysozyme_pxylene.res.fasta', 'w') as wf:
    wf.write(fasta_str)

## topology tables

# Bonds

# you can get a table using mdtraj, but we just use the bonds here
mdj_atoms_df, bonds = traj.top.to_dataframe()

# just the first two columns (atom indices) for our purposes
bonds = bonds[:,0:2]

# we can just write this multiple ways with numpy
np.savetxt('outputs/lysozyme_pxylene.bonds.npy_txt', bonds)
np.save('outputs/lysozyme_pxylene.bonds.npy', bonds)

# make a pandas data frame
bond_df = pd.DataFrame(bonds)

# but wepy provides the ability to get normalized versions for each
# level
chain_df = json_top_chain_df(json_top)
residue_df = json_top_residue_df(json_top)
atom_df = json_top_atom_df(json_top)

bond_df.to_csv('outputs/lysozyme_pxylene.bond.csv', index=False)
chain_df.to_csv('outputs/lysozyme_pxylene.chain.csv', index=False)
residue_df.to_csv('outputs/lysozyme_pxylene.residue.csv', index=False)
atom_df.to_csv('outputs/lysozyme_pxylene.atom.csv', index=False)


# to an SQLite3 database
db = sqlite3.Connection("outputs/lysozyme_pxylene.sqlite3")

bond_df.to_sql('bonds', db)
chain_df.to_sql('chains', db)
residue_df.to_sql('residues', db)
atom_df.to_sql('atoms', db)

# to an HDF5 file
bond_df.to_hdf('outputs/lysozyme_pxylene.top.h5', 'bonds')
chain_df.to_hdf('outputs/lysozyme_pxylene.top.h5', 'chains')
residue_df.to_hdf('outputs/lysozyme_pxylene.top.h5', 'residues')
atom_df.to_hdf('outputs/lysozyme_pxylene.top.h5', 'atoms')

# to an excel spreadsheet
with pd.ExcelWriter('outputs/lysozyme_pxylene.top.xlsx', mode='r+') as writer:

    bond_df.to_excel(writer, sheet_name='bonds')
    chain_df.to_excel(writer, sheet_name='chains')
    residue_df.to_excel(writer, sheet_name='residues')
    atom_df.to_excel(writer, sheet_name='atoms')


## coordinates

# separately, in binary format
coords = traj.xyz

np.savez('outputs/lysozyme_pxylene_reference.npz', coords)



import numpy as np
import pytest
from geomm.distance import minimum_distance

def test_minimum_distance_simple():
    coordsA = np.array([[0, 0, 0], [1, 1, 1]])
    coordsB = np.array([[2, 2, 2], [3, 3, 3]])
    # Closest: [1,1,1] to [2,2,2] -> sqrt(3)
    expected = np.linalg.norm(np.array([1,1,1]) - np.array([2,2,2]))
    result = minimum_distance(coordsA, coordsB)
    assert np.isclose(result, expected)

def test_minimum_distance_identical_points():
    coordsA = np.array([[0, 0, 0], [1, 1, 1]])
    coordsB = np.array([[1, 1, 1], [2, 2, 2]])
    # [1,1,1] is in both, so min distance is 0
    assert minimum_distance(coordsA, coordsB) == 0.0

def test_minimum_distance_multiple_minima():
    coordsA = np.array([[0, 0, 0], [2, 2, 2]])
    coordsB = np.array([[1, 1, 1], [1, 1, 1]])
    # Both points in B are equidistant to both in A
    expected = np.linalg.norm(np.array([1,1,1]) - np.array([0,0,0]))
    result = minimum_distance(coordsA, coordsB)
    assert np.isclose(result, expected)

def test_minimum_distance_assertion_error():
    coordsA = np.array([[0, 0], [1, 1]])  # Not 3D
    coordsB = np.array([[2, 2], [3, 3]])
    with pytest.raises(AssertionError):
        minimum_distance(coordsA, coordsB)


import numpy as np

def centroid(coords, weights=None):
    """Return the centroid (AKA center of geometry or when weights
    correspond to masses the 'center of mass (COM)') of a set of
    coordinates.

    Parameters
    ----------

    coords : arraylike
        The coordinates to find the centroid of.

    weights : arraylike, optional
        Optional weights for the coordinates to compute a weighted
        centroid (center of mass).
       (Default = None)

    """

    # no weights
    if weights is None:
        centroid = np.average(coords, axis=0)
    # use the weights
    else:
        centroid = np.average(coords, axis=0, weights=weights)

    return centroid



import numpy as np

from geomm.theobald_qcp import theobald_qcp
from geomm.centering import center
from geomm.centroid import centroid

def superimpose(ref_coords, coords, idxs=None, weights=None):
    """Superimpose a set of coordinates to reference coordinates using the
    Theobald-QCP method.

    Assumes coordinates are centered appropriately.

    Parameters
    ----------

    ref_coords : arraylike
        The template coordinates that `coords` will be aligned to.

    coords : arraylike
        The coordinates which will be aligned to the template
        coordinates and have the rotation matrix applied to them and
        returned.

    idxs : arraylike, optional
        If given will superimpose the coordinates based only on the
        alignment on this subset of atoms to the reference.
       (Default = None)

    weights : arraylike, optional
        Give weights to the coordinates for a weighted centroid
        ('center of mass') used in translations to the origin and to
        the centroid of the reference structure.
       (Default = None)

    Returns
    -------

    superimposed_coords : arraylike of float
        The transformed coordinates.

    rotation_matrix : arraylike
        The rotation matrix from Theobald-QCP that minimized the RMSD

    qcp_rmsd : float
        The RMSD calculated from Theobald-QCP.

    Warnings
    --------

    Assumes that the coordinates are both centered.

    """

    # find the centroid of the reference coordinates
    if idxs is not None:
        ref_centroid = centroid(ref_coords[idxs], weights=weights)
    else:
        ref_centroid = centroid(ref_coords, weights=weights)

    # perform the theobald_qcp method to get the rotation matrix
    qcp_rmsd, rotation_matrix = theobald_qcp(ref_coords, coords,
                                             idxs=idxs, weights=weights)

    # rotate coords according to the rotation matrix
    rot_coords = np.dot(coords, rotation_matrix)

    # translate the rotated coordinates to the reference centroid
    sup_coords = rot_coords + ref_centroid

    return sup_coords, rotation_matrix, qcp_rmsd

# the following method contains portions of the software mdtraj which
# is distributed under the following license
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Peter Eastman, Robert McGibbon
# Contributors: Kyle A. Beauchamp, Matthew Harrigan, Carlos Xavier Hernandez
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit, copyright (c) 2012 Stanford University and Peter Eastman. Those
# portions are distributed under the following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################

def alt_superimpose(ref_coords, coords):
    """Returns the translation and rotation mapping mobile onto target.

    Parameters
    ----------

    mobile : ndarray, shape = (n_atoms, 3)
        xyz coordinates of a `single` frame, to be aligned onto target.

    target : ndarray, shape = (n_atoms, 3)
        xyz coordinates of a `single` frame

    Returns
    -------

    translation : ndarray, shape=(3,)
        Difference between the centroids of the two conformations

    rotation : ndarray, shape=(3,3)
        Rotation matrix to apply to mobile to carry out the transformation.

    """

    # ensure_type(mobile, 'float', 2, 'mobile', warn_on_cast=False, shape=(None, 3))
    # ensure_type(target, 'float', 2, 'target', warn_on_cast=False, shape=(target.shape[0], 3))

    mu1 = ref_coords.mean(0)
    mu2 = coords.mean(0)

    ref_coords = ref_coords - mu1
    coords = coords - mu2

    correlation_matrix = np.dot(np.transpose(ref_coords), coords)
    V, S, W_tr = np.linalg.svd(correlation_matrix)
    is_reflection = (np.linalg.det(V) * np.linalg.det(W_tr)) < 0.0
    if is_reflection:
        V[:, -1] = -V[:, -1]
    rotation = np.dot(V, W_tr)

    translation = mu2 - mu1.dot(rotation)

    return translation, rotation



#!/usr/bin/env python


import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import setup, find_packages

import itertools as it

# package specific imports
from Cython.Build import cythonize
import numpy as np

# the basic needed requirements for a package
base_requirements = [
    'numpy',
    'scipy',
    'pint'
]

# extras requirements list

# SNIPPET: example extra requirement
# example_extra_requirements = ['requests']
# extras = [example_extra_requirements,]

# Add your extra requirements lists here:
extras = [
]

# combination of all the extras requirements
_all_requirements = [[base_requirements]] + extras
all_requirements = it.chain.from_iterable(_all_requirements)

setup(
    name='geomm',
    version='0.3',
    author="Samuel D. Lotz",
    author_email="samuel.lotz@salotz.info",
    description="A simple no-nonsense library for computing common geometry on macromolecular systems.",
    #long_description=open('README.org').read(),
    license="MIT",
    url="https://github.com/ADicksonLab/geomm",
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3'
    ],

    keywords='geometry chemistry macromolecules protein structural-informatics',
    # building/dev
    setup_requires=[
        'pytest-runner',
        'numpy',
        'cython',
    ],
    tests_require=['pytest', 'tox'],

    include_dirs=[np.get_include()],
    ext_modules = cythonize("src/geomm/pyqcprot.pyx"),


    # package
    packages=find_packages(where='src'),

    package_dir={'' : 'src'},

    # if this is true then the package_data won't be included in the
    # dist. Use MANIFEST.in for this
    include_package_data=True,

    # pymodules is for single file standalone modules not part of the
    # package
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],

    install_requires=base_requirements,

    # SNIPPET: example of using extra requirements
    # extras_require={
    #     'extras' : example_extra_requirements
    #     'all' : all_requirements,
    # }

    # include your extra requirement sets here
    extras_require={
        'all' : all_requirements,
    }
)



import numpy as np
import pytest
from geomm.rmsd import calc_rmsd

def test_rmsd_identical_coords():
    ref = np.array([[0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0]])
    coords = np.copy(ref)
    assert calc_rmsd(ref, coords) == 0.0

def test_rmsd_simple_translation():
    ref = np.array([[0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0]])
    coords = ref + 1.0
    expected = np.sqrt(np.sum((coords - ref)**2) / ref.shape[0])
    assert np.isclose(calc_rmsd(ref, coords), expected)

def test_rmsd_with_idxs():
    ref = np.array([[0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0]])
    coords = np.array([[0.0, 0.0, 0.0],
                       [2.0, 0.0, 0.0],
                       [0.0, 2.0, 0.0]])
    idxs = np.array([1, 2])
    expected = np.sqrt(np.sum((coords[idxs] - ref[idxs])**2) / idxs.shape[0])
    assert np.isclose(calc_rmsd(ref, coords, idxs), expected)

def test_rmsd_shape_assertion():
    ref = np.zeros((3, 3))
    coords = np.zeros((4, 3))
    with pytest.raises(AssertionError):
        calc_rmsd(ref, coords)

def test_rmsd_dimension_assertion():
    ref = np.zeros((3, 2))
    coords = np.zeros((3, 2))
    with pytest.raises(AssertionError):
        calc_rmsd(ref, coords)

def test_rmsd_rank_assertion():
    ref = np.zeros((3, 3))
    coords = np.zeros((3, 3, 1))
    with pytest.raises(AssertionError):
        calc_rmsd(ref, coords)


import numpy as np

from geomm.pyqcprot import CalcRMSDRotationalMatrix

def theobald_qcp(ref_coords, coords, idxs=None, weights=None):
    """Wrapper around the pyqcprot implementation of the Theobald-QCP
    method for the calculation of RMSD and the RMSD minimizing rotation
    matrix. This function just gives a more pythonic API to the
    function

    Parameters
    ----------

    ref_coords : arraylike
        The refence coordinates that will be aligned to.

    coords : arraylike
        The coordinates that will be rotated to match ref_coords.

    idxs : arraylike of int
        Indices of the atoms that you want to align. Rotation will be
        applied to all coordinates.

    weights : arraylike, optional
        If your coordinates are weighted (e.g. mass) this is an
        array of those weights
       (Default = None)

    Returns
    -------

    rmsd : float
        The rmsd of the two sets of coordinates.

    rotation_matrix : arraylike
        The rotation matrix that minimizes the RMSD.

    """

    # make sure the coords are the same size
    assert ref_coords.shape[0] == coords.shape[0], \
        "Number of coordinates are not the same"

    # make sure the number of dimensions is 3
    assert (ref_coords.shape[1] == 3) and (coords.shape[1] == 3), \
        "Number of dimensions are not the same"

    # copy the coordinates to make sure we are not modifying them in
    # place
    ref_coords_copy = np.array(ref_coords.copy(), dtype=np.float64)
    coords_copy = np.array(coords.copy(), dtype=np.float64)

    # if idxs were given we use just those for aligning
    if idxs is not None:
        align_ref_coords = ref_coords_copy[idxs]
        align_coords = coords_copy[idxs]
    else:
        align_ref_coords = ref_coords_copy
        align_coords = coords_copy

    # the number of coordinates (atoms)
    n_coords = align_ref_coords.shape[0]

    # make sure the weights if given are the right size
    if weights is not None:
        assert weights.shape[0] == n_coords, \
            "Number of weights given does not match the number of coordinates"

    # make an empty rotation matrix to fill from the cython function,
    # this is always an array of length 9
    rotation_matrix = np.zeros((9,), dtype=np.float64)

    # calculate the rmsd, and the rotation matrix which is modified in-place
    rmsd = CalcRMSDRotationalMatrix(align_ref_coords, align_coords,
                                    n_coords, rotation_matrix, weights)

    # reshape the rotation matrix to be 2D
    return rmsd, rotation_matrix.reshape( (3, 3) )
